import itertools

from dune.perftool.pdelab.argument import (name_accumulation_variable,
                                           PDELabAccumulationFunction,
                                           )
from dune.perftool.generation import (backend,
                                      domain,
                                      dump_accumulate_timer,
                                      get_counted_variable,
                                      get_counter,
                                      iname,
                                      instruction,
                                      post_include,
                                      kernel_cached,
                                      temporary_variable,
                                      transform,
                                      )
from dune.perftool.options import get_option
from dune.perftool.loopy.flatten import flatten_index
from dune.perftool.loopy.buffer import get_buffer_temporary
from dune.perftool.sumfact.quadrature import nest_quadrature_loops
from dune.perftool.pdelab.localoperator import determine_accumulation_space
from dune.perftool.pdelab.restriction import restricted_name
from dune.perftool.pdelab.signatures import assembler_routine_name
from dune.perftool.pdelab.geometry import world_dimension
from dune.perftool.sumfact.tabulation import (basis_functions_per_direction,
                                              construct_basis_matrix_sequence,
                                              )
from dune.perftool.sumfact.switch import (get_facedir,
                                          get_facemod,
                                          )
from dune.perftool.sumfact.symbolic import SumfactKernel, SumfactKernelInputBase
from dune.perftool.ufl.modified_terminals import extract_modified_arguments
from dune.perftool.tools import get_pymbolic_basename
from dune.perftool.error import PerftoolError
from dune.perftool.sumfact.quadrature import quadrature_inames

from pytools import ImmutableRecord

import loopy as lp
import numpy as np
import pymbolic.primitives as prim
import ufl.classes as uc
from ufl import FiniteElement, TensorProductElement


@iname
def _sumfact_iname(bound, _type, count):
    name = "sf_{}_{}".format(_type, str(count))
    domain(name, bound)
    return name


def sumfact_iname(bound, _type):
    count = get_counter('_sumfac_iname_{}'.format(_type))
    name = _sumfact_iname(bound, _type, count)
    return name


@kernel_cached
def accum_iname(element, bound, i):
    if element is None:
        suffix = ""
    else:
        from dune.perftool.pdelab.driver import FEM_name_mangling
        suffix = "_{}".format(FEM_name_mangling(element))
    return sumfact_iname(bound, "accum{}".format(suffix))


class AlreadyAssembledInput(SumfactKernelInputBase):
    def __init__(self, index):
        self.index = index

    def __eq__(self, other):
        return type(self) == type(other) and self.index == other.index

    def __repr__(self):
        return "AlreadyAssembledInput({})".format(self.index)

    def __hash__(self):
        return hash(self.index)


class SumfactAccumulationInfo(ImmutableRecord):
    def __init__(self,
                 element=None,
                 element_index=0,
                 restriction=None,
                 inames=(),
                 grad_index=None,
                 ):
        ImmutableRecord.__init__(self,
                                 element=element,
                                 element_index=element_index,
                                 restriction=restriction,
                                 inames=inames,
                                 grad_index=grad_index,
                                 )

    def __eq__(self, other):
        return type(self) == type(other) and self.element_index == other.element_index and self.restriction == other.restriction and self.grad_index == other.grad_index

    def __hash__(self):
        return (self.element_index, self.restriction, self.grad_index)


def get_accumulation_info(expr, visitor):
    element = expr.ufl_element()
    leaf_element = element
    element_index = 0
    from ufl import MixedElement
    if isinstance(expr.ufl_element(), MixedElement):
        element_index = visitor.indices[0]
        leaf_element = element.extract_component(element_index)[1]

    restriction = visitor.restriction
    if visitor.measure == 'exterior_facet':
        from dune.perftool.pdelab.restriction import Restriction
        restriction = Restriction.NEGATIVE

    inames = visitor.interface.lfs_inames(leaf_element,
                                          restriction,
                                          expr.number()
                                          )

    grad_index = None
    if visitor.reference_grad and expr.number() == 0:
        if isinstance(expr.ufl_element(), MixedElement):
            grad_index = visitor.indices[1]
        else:
            grad_index = visitor.indices[0]

    return SumfactAccumulationInfo(element=expr.ufl_element(),
                                   element_index=element_index,
                                   restriction=restriction,
                                   inames=inames,
                                   grad_index=grad_index,
                                   )


def _get_childs(element):
    if isinstance(element, (FiniteElement, TensorProductElement)):
        yield (0, element)
    else:
        for i in range(element.value_size()):
            yield (i, element.extract_component(i)[1])


def _test_generator(expr, visitor):
    from dune.perftool.ufl.modified_terminals import extract_modified_arguments
    ma = extract_modified_arguments(expr, argnumber=0)
    if len(ma) == 0:
        return
    element = ma[0].argexpr.ufl_element()
    dim = world_dimension()

    from dune.perftool.ufl.modified_terminals import Restriction
    if visitor.measure == "cell":
        restrictions = (Restriction.NONE,)
    elif visitor.measure == "exterior_facet":
        restrictions = (Restriction.NEGATIVE,)
    elif visitor.measure == "interior_facet":
        restrictions = (Restriction.NEGATIVE, Restriction.POSITIVE)
    for res in restrictions:
        for ei, e in _get_childs(element):
            for grad in (None,) + tuple(range(dim)):
                yield SumfactAccumulationInfo(element_index=ei,
                                              restriction=res,
                                              grad_index=grad)


def _trial_generator(expr, visitor):
    from dune.perftool.ufl.modified_terminals import extract_modified_arguments
    ma = extract_modified_arguments(expr, argnumber=1)
    if len(ma) == 0:
        yield None
        return
    element = ma[0].argexpr.ufl_element()

    from dune.perftool.ufl.modified_terminals import Restriction
    if visitor.measure == "cell":
        restrictions = (Restriction.NONE,)
    elif visitor.measure == "exterior_facet":
        restrictions = (Restriction.NEGATIVE,)
    elif visitor.measure == "interior_facet":
        restrictions = (Restriction.NEGATIVE, Restriction.POSITIVE)
    for res in restrictions:
        for ei, e in _get_childs(element):
            yield SumfactAccumulationInfo(element_index=ei, restriction=res, element=e)


def list_accumulation_infos(expr, visitor):
    import itertools
    return itertools.product(_test_generator(expr, visitor), _trial_generator(expr, visitor))


def generate_accumulation_instruction(expr, visitor):
    dim = world_dimension()
    test_info = visitor.test_info
    trial_info = visitor.trial_info

    # Number of basis functions per direction
    leaf_element = test_info.element
    if leaf_element.num_sub_elements() > 0 and not isinstance(leaf_element, TensorProductElement):
        leaf_element = leaf_element.extract_component(test_info.element_index)[1]
    degree = leaf_element._degree
    if isinstance(degree, int):
        degree = (degree,) * dim
    basis_size = tuple(deg + 1 for deg in degree)

    # Anisotropic finite elements are not (yet) supported by Dune
    assert(size == basis_size[0] for size in basis_size)

    from dune.perftool.pdelab.localoperator import boundary_predicates
    predicates = boundary_predicates(expr,
                                     visitor.measure,
                                     visitor.subdomain_id,
                                     visitor)

    insn_dep = None

    from dune.perftool.pdelab.localoperator import determine_accumulation_space
    test_lfs = determine_accumulation_space(test_info, 0)
    ansatz_lfs = determine_accumulation_space(trial_info, 1)

    if trial_info is None:
        trial_info = SumfactAccumulationInfo()
    trial_leaf_element = trial_info.element
    from ufl import MixedElement
    if isinstance(trial_leaf_element, MixedElement):
        trial_leaf_element = trial_leaf_element.extract_component(trial_info.element_index)[1]

    from dune.perftool.pdelab.argument import name_accumulation_variable
    accumvar = name_accumulation_variable(test_lfs.get_restriction() + ansatz_lfs.get_restriction())

    matrix_sequence = construct_basis_matrix_sequence(
        transpose=True,
        derivative=test_info.grad_index,
        facedir=get_facedir(test_info.restriction),
        facemod=get_facemod(test_info.restriction),
        basis_size=basis_size)

    jacobian_inames = trial_info.inames
    priority = test_info.grad_index
    if priority is None:
        priority = 3

    sf = SumfactKernel(matrix_sequence=matrix_sequence,
                       restriction=(test_info.restriction, trial_info.restriction),
                       stage=3,
                       position_priority=priority,
                       accumvar=accumvar,
                       trial_element=trial_leaf_element,
                       input=AlreadyAssembledInput(index=(test_info.element_index,)),
                       component_index=test_info.element_index,
                       )

    from dune.perftool.sumfact.vectorization import attach_vectorization_info
    vsf = attach_vectorization_info(sf)

    # Make sure we have a buffer that we can set up the input with
    buffer = vsf.buffer
    if buffer is None:
        buffer = get_counted_variable("buffer")

    vectag = frozenset({"gradvec"}) if vsf.vectorized else frozenset()

    temp = get_buffer_temporary(buffer,
                                shape=vsf.quadrature_shape,
                                dim_tags=vsf.quadrature_dimtags,
                                name="input_{}".format(buffer),
                                )

    # Those input fields, that are padded need to be set to zero
    # in order to do a horizontal_add later on
    for pad in vsf.padded_indices:
        assignee = prim.Subscript(lp.TaggedVariable(temp, vsf.tag), pad)
        instruction(assignee=assignee,
                    expression=0,
                    forced_iname_deps=frozenset(quadrature_inames(trial_leaf_element) + jacobian_inames),
                    forced_iname_deps_is_final=True,
                    tags=frozenset(["quadvec", "gradvec"]),
                    )

    # Write timing stuff for jacobian (for alpha methods it is done at the end of stage 1)
    timer_dep = frozenset()
    if get_option("instrumentation_level") >= 4:
        timer_name = assembler_routine_name() + '_kernel' + '_quadratureloop'
        post_include('HP_DECLARE_TIMER({});'.format(timer_name), filetag='operatorfile')
        dump_accumulate_timer(timer_name)
        if(jacobian_inames):
            timer_dep = frozenset({instruction(code="HP_TIMER_START({});".format(timer_name),
                                               within_inames=frozenset(jacobian_inames))})

    # Determine dependencies
    from loopy.match import Or, Writes
    from loopy.symbolic import DependencyMapper
    from dune.perftool.tools import get_pymbolic_basename
    deps = Or(tuple(Writes(get_pymbolic_basename(e)) for e in DependencyMapper()(expr)))

    # Issue an instruction in the quadrature loop that fills the buffer
    # with the evaluation of the contribution at all quadrature points
    assignee = prim.Subscript(lp.TaggedVariable(temp, vsf.tag),
                              vsf.quadrature_index(sf))
    contrib_dep = instruction(assignee=assignee,
                              expression=expr,
                              forced_iname_deps=frozenset(quadrature_inames(trial_leaf_element) + jacobian_inames),
                              forced_iname_deps_is_final=True,
                              tags=frozenset({"quadvec"}).union(vectag),
                              depends_on=frozenset({deps}).union(timer_dep).union(frozenset({lp.match.Tagged("sumfact_stage1")})),
                              )

    if insn_dep is None:
        insn_dep = frozenset({contrib_dep})

    if get_option("instrumentation_level") >= 4:
        insn_dep = frozenset({instruction(code="HP_TIMER_STOP({});".format(timer_name),
                                          depends_on=insn_dep,
                                          within_inames=frozenset(jacobian_inames))})

    inames = tuple(accum_iname(trial_leaf_element, mat.rows, i)
                   for i, mat in enumerate(vsf.matrix_sequence))

    # Collect the lfs and lfs indices for the accumulate call
    test_lfs.index = flatten_index(tuple(prim.Variable(i) for i in inames),
                                   basis_size,
                                   order="f"
                                   )

    # In the jacobian case, also determine the space for the ansatz space
    if jacobian_inames:
        # TODO the next line should get its inames from
        # elsewhere. This is *NOT* robust (but works right now)
        from dune.perftool.sumfact.basis import _basis_functions_per_direction
        ansatz_lfs.index = flatten_index(tuple(prim.Variable(jacobian_inames[i])
                                               for i in range(world_dimension())),
                                         _basis_functions_per_direction(trial_leaf_element),
                                         order="f"
                                         )

    # Add a sum factorization kernel that implements the multiplication
    # with the test function (stage 3)
    from dune.perftool.sumfact.realization import realize_sum_factorization_kernel
    result, insn_dep = realize_sum_factorization_kernel(vsf.copy(insn_dep=vsf.insn_dep.union(insn_dep)))

    # Determine the expression to accumulate with. This depends on the vectorization strategy!
    result = prim.Subscript(result, tuple(prim.Variable(i) for i in inames))
    vecinames = ()

    if vsf.vectorized:
        iname = accum_iname(trial_leaf_element, vsf.vector_width, "vec")
        vecinames = (iname,)
        transform(lp.tag_inames, [(iname, "vec")])
        from dune.perftool.tools import maybe_wrap_subscript
        result = prim.Call(prim.Variable("horizontal_add"),
                           (maybe_wrap_subscript(result, prim.Variable(iname)),),
                           )

    if not get_option("fastdg"):
        rank = 2 if jacobian_inames else 1
        expr = prim.Call(PDELabAccumulationFunction(accumvar, rank),
                         (test_lfs.get_args() +
                          ansatz_lfs.get_args() +
                          (result,)
                          )
                         )
        instruction(assignees=(),
                    expression=expr,
                    forced_iname_deps=frozenset(inames + vecinames + jacobian_inames),
                    forced_iname_deps_is_final=True,
                    depends_on=insn_dep,
                    predicates=predicates
                    )
